/*
 * Copyright (C) 2015 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#ifndef B3Value_h
#define B3Value_h

#if ENABLE(B3_JIT)

#include "AirArg.h"
#include "B3Effects.h"
#include "B3Opcode.h"
#include "B3Origin.h"
#include "B3Type.h"
#include <wtf/FastMalloc.h>
#include <wtf/Noncopyable.h>

namespace JSC { namespace B3 {

class BasicBlock;
class Procedure;
class Stackmap;

class JS_EXPORT_PRIVATE Value {
    WTF_MAKE_NONCOPYABLE(Value);
    WTF_MAKE_FAST_ALLOCATED;
public:
    typedef Vector<Value*, 3> AdjacencyList;

    static const char* const dumpPrefix;

    static bool accepts(Opcode) { return true; }

    virtual ~Value();

    unsigned index() const { return m_index; }
    
    // Note that the opcode is immutable, except for replacing values with Identity or Nop.
    Opcode opcode() const { return m_opcode; }

    Origin origin() const { return m_origin; }
    
    Value*& child(unsigned index) { return m_children[index]; }
    Value* child(unsigned index) const { return m_children[index]; }

    Value*& lastChild() { return m_children.last(); }
    Value* lastChild() const { return m_children.last(); }

    unsigned numChildren() const { return m_children.size(); }

    // This computes the type using the opcode.
    Type type() const { return m_type; }

    // This is useful when lowering. Note that this is only valid for non-void values.
    Air::Arg::Type airType() const { return Air::Arg::typeForB3Type(type()); }

    AdjacencyList& children() { return m_children; } 
    const AdjacencyList& children() const { return m_children; }

    void replaceWithIdentity(Value*);
    void replaceWithNop();

    void dump(PrintStream&) const;
    void deepDump(PrintStream&) const;

    // This is how you cast Values. For example, if you want to do something provided that we have a
    // ArgumentRegValue, you can do:
    //
    // if (ArgumentRegValue* argumentReg = value->as<ArgumentRegValue>()) {
    //     things
    // }
    //
    // This will return null if this opcode() != ArgumentReg. This works because this returns nullptr
    // if T::accepts(opcode()) returns false.
    template<typename T>
    T* as();
    template<typename T>
    const T* as() const;

    // What follows are a bunch of helpers for inspecting and modifying values. Note that we have a
    // bunch of different idioms for implementing such helpers. You can use virtual methods, and
    // override from the various Value subclasses. You can put the method inside Value and make it
    // non-virtual, and the implementation can switch on opcode. The method could be inline or not.
    // If a method is specific to some Value subclass, you could put it in the subclass, or you could
    // put it on Value anyway. It's fine to pick whatever feels right, and we shouldn't restrict
    // ourselves to any particular idiom.

    bool isConstant() const;
    
    virtual Value* negConstant(Procedure&) const;
    virtual Value* addConstant(Procedure&, int32_t other) const;
    virtual Value* addConstant(Procedure&, Value* other) const;
    virtual Value* subConstant(Procedure&, Value* other) const;
    virtual Value* equalConstant(Procedure&, Value* other) const;
    virtual Value* notEqualConstant(Procedure&, Value* other) const;

    bool hasInt32() const;
    int32_t asInt32() const;
    bool isInt32(int32_t) const;
    
    bool hasInt64() const;
    int64_t asInt64() const;
    bool isInt64(int64_t) const;

    bool hasInt() const;
    int64_t asInt() const;
    bool isInt(int64_t value) const;

    bool hasIntPtr() const;
    intptr_t asIntPtr() const;
    bool isIntPtr(intptr_t) const;

    bool hasDouble() const;
    double asDouble() const;
    bool isEqualToDouble(double) const; // We say "isEqualToDouble" because "isDouble" would be a bit equality.

    bool hasNumber() const;
    template<typename T> bool representableAs() const;
    template<typename T> T asNumber() const;

    // Booleans in B3 are Const32(0) or Const32(1). So this is true if the type is Int32 and the only
    // possible return values are 0 or 1. It's OK for this method to conservatively return false.
    bool returnsBool() const;

    TriState asTriState() const;
    bool isLikeZero() const { return asTriState() == FalseTriState; }
    bool isLikeNonZero() const { return asTriState() == TrueTriState; }
    
    Effects effects() const;

    Stackmap* stackmap();

    // Makes sure that none of the children are Identity's. If a child points to Identity, this will
    // repoint it at the Identity's child. For simplicity, this will follow arbitrarily long chains
    // of Identity's.
    void performSubstitution();

protected:
    virtual void dumpMeta(PrintStream&) const;
    
private:
    friend class Procedure;

protected:
    // Instantiate values via Procedure.
    // This form requires specifying the type explicitly:
    template<typename... Arguments>
    explicit Value(unsigned index, Opcode opcode, Type type, Origin origin, Value* firstChild, Arguments... arguments)
        : m_index(index)
        , m_opcode(opcode)
        , m_type(type)
        , m_origin(origin)
        , m_children{ firstChild, arguments... }
    {
    }
    // This form is for specifying the type explicitly when the opcode has no children:
    explicit Value(unsigned index, Opcode opcode, Type type, Origin origin)
        : m_index(index)
        , m_opcode(opcode)
        , m_type(type)
        , m_origin(origin)
    {
    }
    // This form is for those opcodes that can infer their type from the opcode and first child:
    template<typename... Arguments>
    explicit Value(unsigned index, Opcode opcode, Origin origin, Value* firstChild, Arguments... arguments)
        : m_index(index)
        , m_opcode(opcode)
        , m_type(typeFor(opcode, firstChild))
        , m_origin(origin)
        , m_children{ firstChild, arguments... }
    {
    }
    // This form is for those opcodes that can infer their type from the opcode alone, and that don't
    // take any arguments:
    explicit Value(unsigned index, Opcode opcode, Origin origin)
        : m_index(index)
        , m_opcode(opcode)
        , m_type(typeFor(opcode, nullptr))
        , m_origin(origin)
    {
    }
    // Use this form for varargs.
    explicit Value(unsigned index, Opcode opcode, Type type, Origin origin, const AdjacencyList& children)
        : m_index(index)
        , m_opcode(opcode)
        , m_type(type)
        , m_origin(origin)
        , m_children(children)
    {
    }
    explicit Value(unsigned index, Opcode opcode, Type type, Origin origin, AdjacencyList&& children)
        : m_index(index)
        , m_opcode(opcode)
        , m_type(type)
        , m_origin(origin)
        , m_children(WTF::move(children))
    {
    }

private:
    static Type typeFor(Opcode, Value* firstChild);

    // This group of fields is arranged to fit in 64 bits.
    unsigned m_index;
    Opcode m_opcode;
    Type m_type;
    
    Origin m_origin;
    AdjacencyList m_children;

public:
    BasicBlock* owner { nullptr }; // computed by Procedure::resetValueOwners().
};

class DeepValueDump {
public:
    DeepValueDump(const Value* value)
        : m_value(value)
    {
    }

    void dump(PrintStream& out) const
    {
        if (m_value)
            m_value->deepDump(out);
        else
            out.print("<null>");
    }

private:
    const Value* m_value;
};

inline DeepValueDump deepDump(const Value* value)
{
    return DeepValueDump(value);
}

} } // namespace JSC::B3

#endif // ENABLE(B3_JIT)

#endif // B3Value_h

